#include "mex.h"
#include <stdio.h>
#include <math.h>

#ifndef max
#define max( a, b ) ( ((a) > (b)) ? (a) : (b) )
#endif

#ifndef min
#define min( a, b ) ( ((a) < (b)) ? (a) : (b) )
#endif

typedef struct {
	int numIndiceAccum;
	int numIter;
	int numIndice;
	int *indice;
	
} indice;

typedef	struct {
	indice *diffIndice;
	double **overlapMatrix;
	
} reUseStruct;


/* Prototypes*/
void calcSegSysCovarianceRow(int numSamps, int numAggSegs, int segI, int numCol, int **lrSegI, double **segStarts, double **segEnds, reUseStruct reuseVar, double *covRow);
void getDiffIndice(int numRows, int segJ, int **matrix, indice *diffIndice);
double sumMatrixElem(int numRows, int numCols, double **matrix);
double calcOverlap(int segI, int segJ, int i, int j, int **lrSegI, double **segStarts, double **segEnds);
int **mexDoubleMatrix2IntMatrix(int numRows, int numCols, double *mexMatrix);
void freeIntMatrix(int numRows, int **mexMatrix);
double **mexDoubleMatrix2DoubleMatrix(int numRows, int numCols, double *mexMatrix);
double **createDoubleMatrix(int numRows, int numCols);
void freeDoubleMatrix(int numRows, double **mexMatrix);




void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
    
	int segI, numCol, numSamps, numAggSegs, maxSegsInSamp;
	int i;
	double *lrSegIDouble; 
	double *segStartsMx, *segEndsMx;
	double *covRow, *covRowTemp, *covMatrix;
	double **segStarts, **segEnds;
	int **lrSegI;
	reUseStruct reuseVar;
	   
/*	Read input parameters */
	segI			= (int)(*mxGetPr(prhs[0]));
	segI			= segI - 1;
	numCol			= (int)(*mxGetPr(prhs[1]));
	numSamps		= mxGetM(prhs[2]);
	numAggSegs		= mxGetN(prhs[2]);
	lrSegIDouble	= mxGetPr(prhs[2]);
	lrSegI			= mexDoubleMatrix2IntMatrix(numSamps, numAggSegs, lrSegIDouble);
	maxSegsInSamp	= mxGetN(prhs[3]);
	segStartsMx		= mxGetPr(prhs[3]);
	segEndsMx		= mxGetPr(prhs[4]);
	segStarts		= mexDoubleMatrix2DoubleMatrix(numSamps, maxSegsInSamp, segStartsMx);
	segEnds			= mexDoubleMatrix2DoubleMatrix(numSamps, maxSegsInSamp, segEndsMx);
	
	plhs[0]			= mxCreateDoubleMatrix(numCol, numAggSegs, mxREAL);
	covMatrix		= mxGetPr(plhs[0]);
	
/*	printf("segI = %d\n", segI);
	printf("numCol = %d\n", numCol);
	printf("numSamps = %d; numAggSegs = %d\n", numSamps, numAggSegs);
	printf("2X2 elements of lrSegI = %d, %d; %d, %d\n", lrSegI[99][99], lrSegI[99][100], lrSegI[100][99], lrSegI[100][100]);
	printf("2X2 elements of segStarts = %f, %f; %f, %f\n", segStarts[49][19], segStarts[49][20], segStarts[50][19], segStarts[50][20]);
	printf("2X2 elements of segEnds = %f, %f; %f, %f\n", segEnds[49][19], segEnds[49][20], segEnds[50][19], segEnds[50][20]);*/
	reuseVar.diffIndice = (indice *)malloc(sizeof(indice));
	reuseVar.diffIndice->numIndiceAccum = 0;
	reuseVar.diffIndice->numIter = 0;
	reuseVar.diffIndice->indice = (int *)malloc(sizeof(int) * numSamps);
	reuseVar.overlapMatrix = createDoubleMatrix(numSamps, numSamps);
	
	for (segI = 0; segI < numAggSegs; segI++) {
		
		calcSegSysCovarianceRow(numSamps, numAggSegs, segI, numCol, lrSegI, segStarts, segEnds, reuseVar, &(covMatrix[numCol*segI]));
/*		for (i = 0; i < numCol; i++) {
			covMatrix[i] = covRowTemp[numCol*segI + i];
		}
		free(covRowTemp);*/
		if ((segI % 100) == 0) {
			printf("segI = %d\n", segI);
		}
	}
	
	printf("averageNumSegChange = %f\n", reuseVar.diffIndice->numIndiceAccum / reuseVar.diffIndice->numIter);
    
	freeDoubleMatrix(numSamps, reuseVar.overlapMatrix);
	free(reuseVar.diffIndice->indice);
	free(reuseVar.diffIndice);
		
	freeDoubleMatrix(numSamps, segEnds);
	freeDoubleMatrix(numSamps, segStarts);
	freeIntMatrix(numSamps, lrSegI);
    
    return;
}

void calcSegSysCovarianceRow(int numSamps, int numAggSegs, int segI, int numCol, int **lrSegI, double **segStarts, double **segEnds, reUseStruct reuseVar, double *covRow){
	int i, j, k;
	int segJ, covIndex;
	
	/*printf("segI = %d\n", segI);*/
	for (i = 0; i < numSamps; i++) {
		for (j = 0; j < numSamps; j++) {
			reuseVar.overlapMatrix[i][j] = calcOverlap(segI, segI, i, j, lrSegI, segStarts, segEnds);
		}
	}
	covRow[0] = sumMatrixElem(numSamps, numSamps, reuseVar.overlapMatrix);
	
	for (segJ = segI+1, covIndex = 1; segJ < min(segI + numCol, numAggSegs); segJ++, covIndex++) {
		getDiffIndice(numSamps, segJ, lrSegI, reuseVar.diffIndice);
		for (k = 0, i = reuseVar.diffIndice->indice[k]; k < reuseVar.diffIndice->numIndice; k++, i = reuseVar.diffIndice->indice[k]) {
			for (j = 0; j < numSamps; j++) {
				reuseVar.overlapMatrix[i][j] = calcOverlap(segI, segJ, i, j, lrSegI, segStarts, segEnds);
				reuseVar.overlapMatrix[j][i] = calcOverlap(segI, segJ, j, i, lrSegI, segStarts, segEnds);
			}
		}

		covRow[covIndex] = sumMatrixElem(numSamps, numSamps, reuseVar.overlapMatrix);
	}
	
	return;
}

void getDiffIndice(int numRows, int segJ, int **matrix, indice *diffIndice){
	int i, numChanges;
	
	numChanges = 0;
	for (i = 0; i < numRows; i++) {
		if (matrix[i][segJ] != matrix[i][segJ-1]) {
			numChanges = numChanges + 1;
			diffIndice->indice[numChanges-1] = i;
		}
	}
	diffIndice->numIndice = numChanges;
	diffIndice->numIndiceAccum += numChanges;
	diffIndice->numIter++;
}


double sumMatrixElem(int numRows, int numCols, double **matrix){
	int i, j;
	double sumValue;
	
	sumValue = 0;
	for (i = 0; i < numRows; i++) {
		for (j = 0; j < numCols; j++) {
			sumValue += matrix[i][j];
		}
	}
	
	return sumValue;
}

double calcOverlap(int segI, int segJ, int i, int j, int **lrSegI, double **segStarts, double **segEnds){
	double seg1Start, seg1End, seg2Start, seg2End, seg1Len, seg2Len, span, overlap;
	
	seg1Start	= segStarts[i][lrSegI[i][segI]-1];
	seg1End		= segEnds[i][lrSegI[i][segI]-1];
	seg2Start	= segStarts[j][lrSegI[j][segJ]-1];
	seg2End		= segEnds[j][lrSegI[j][segJ]-1];
	
	seg1Len		= seg1End - seg1Start + 1;
	seg2Len		= seg2End - seg2Start + 1;
	
	span		= max(seg1End, seg2End) - min(seg1Start, seg2Start) + 1;
	overlap		= max(0.0, seg1Len + seg2Len - span) / (seg1Len*seg2Len);
	
	return overlap;
}


int **mexDoubleMatrix2IntMatrix(int numRows, int numCols, double *mexMatrix){
	int i, j;
	int **intMatrix;
	
	intMatrix = (int **)malloc(sizeof(int *) * numRows);
	for (i = 0; i < numRows; i++) {
		intMatrix[i] = (int *)malloc(sizeof(int) * numCols);
		for (j = 0; j < numCols; j++) {
			intMatrix[i][j] = (int)mexMatrix[j*numRows + i];
		}
	}
	
	return intMatrix;
	
}

void freeIntMatrix(int numRows, int **mexMatrix){
	int i;
	for (i = 0; i < numRows; i++) {
		free(mexMatrix[i]);
	}
	free(mexMatrix);
}

double **mexDoubleMatrix2DoubleMatrix(int numRows, int numCols, double *mexMatrix){
	int i, j;
	double **doubleMatrix;
	
	doubleMatrix = (double **)malloc(sizeof(double *) * numRows);
	for (i = 0; i < numRows; i++) {
		doubleMatrix[i] = (double *)malloc(sizeof(double) * numCols);
		for (j = 0; j < numCols; j++) {
			doubleMatrix[i][j] = mexMatrix[j*numRows + i];
		}
	}
	return doubleMatrix;
}

double **createDoubleMatrix(int numRows, int numCols){
	int i, j;
	double **matrix;
	
	matrix = (double **)malloc(sizeof(double *) * numRows);
	for (i = 0; i < numRows; i++) {
		matrix[i] = (double *)calloc(numCols, sizeof(double));
	}
	
	return matrix;
}

void freeDoubleMatrix(int numRows, double **mexMatrix){
	int i;
	for (i = 0; i < numRows; i++) {
		free(mexMatrix[i]);
	}
	free(mexMatrix);
}



