#include "mex.h"
#include <stdio.h>
#include <math.h>

#ifndef max
#define max( a, b ) ( ((a) > (b)) ? (a) : (b) )
#endif

#ifndef min
#define min( a, b ) ( ((a) < (b)) ? (a) : (b) )
#endif

typedef struct {
	int rowI;
	int colI;
} index2Dtype;

index2Dtype **makeMatrixIndex(int matrixLen);
int getMaxCover(int numSamps, double *probeRange);
int *mexDoubleVector2IntVector(int numElem, double *vector);
double **mexDoubleMatrix2DoubleMatrix(int numRows, int numCols, double *mexMatrix);
void freeDoubleMatrix(int numRows, double **mexMatrix);
int rnd(double number);

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
    
	long int sampI, probeI, gI, probeStart, probeEnd, segIStart;
	int	   i, j;
    int    numProbes, numSamps, numCovRows, numCovCols, numWeights, segI;
	int *aggSegI;
	double kernelT, kernelOffset, currKernel1Weight, currKernel2Weight;
	double *probeLoc, *profile, *mapLoc, *probeRange, *g, *kernel1, *aggSegIDouble, *sysCovMatrixMx, *kernel2, *ProfileS;
	double **sysCovMatrix;
	double *kernel1Weights, *kernel2Weights, *kernel1CovWeights;
	index2Dtype **covIndex;
	
/*	Read input parameters */
    numProbes		= mxGetN(prhs[0]);
    probeLoc		= mxGetPr(prhs[0]);
	profile			= mxGetPr(prhs[1]);
	numSamps		= mxGetN(prhs[2]);
	mapLoc			= mxGetPr(prhs[2]);
	probeRange		= mxGetPr(prhs[3]);
	g				= mxGetPr(prhs[4]);
	kernel1			= mxGetPr(prhs[5]);
	aggSegIDouble	= mxGetPr(prhs[6]);
	aggSegI			= mexDoubleVector2IntVector(numProbes, aggSegIDouble);
	numCovRows		= mxGetM(prhs[7]);
	numCovCols		= mxGetN(prhs[7]);
	sysCovMatrixMx  = mxGetPr(prhs[7]);
	sysCovMatrix	= mexDoubleMatrix2DoubleMatrix(numCovRows, numCovCols, sysCovMatrixMx);
	kernel2			= mxGetPr(prhs[8]);
	
	kernelT			= (double)(mxGetScalar(prhs[9]));
	kernelOffset	= (double)(mxGetScalar(prhs[10]));
	
/*	Create outout parameters */
	plhs[0]			= mxCreateDoubleMatrix(1, numSamps, mxREAL);
	ProfileS		= mxGetPr(plhs[0]);

	
	/* Main Code */
	covIndex = makeMatrixIndex(numCovCols);
	kernel1Weights = (double *)malloc(sizeof(double) * numCovCols);
	kernel2Weights = (double *)malloc(sizeof(double) * numCovCols);
	kernel1CovWeights = (double *)malloc(sizeof(double) * numCovCols);
	
	for (sampI = 0; sampI < numSamps; sampI++) {
		probeStart = (long int)(rnd(probeRange[2*sampI]-1));
		probeEnd   = (long int)(rnd(probeRange[2*sampI+1]-1));
		segIStart = aggSegI[probeStart] - 1;
		
		segI = 0;
		currKernel1Weight = 0.0;
		currKernel2Weight = 0.0;
		for (probeI = probeStart, i = 0; probeI <= probeEnd; probeI++, i++) {
			gI = (long int)(rnd((probeLoc[probeI] -  mapLoc[sampI] + kernelOffset) / kernelT));
			currKernel1Weight += kernel1[gI] * profile[probeI];
			currKernel2Weight += kernel2[gI] * profile[probeI];

			if ((probeI == probeEnd) || (aggSegI[probeI] != aggSegI[probeI+1])){
				kernel1Weights[segI] = currKernel1Weight;
				kernel2Weights[segI] = currKernel2Weight;
				segI++;
				currKernel1Weight = 0.0;
				currKernel2Weight = 0.0;
			}
		}
		numWeights = segI;
		/*printf("numWeights = %d\n", numWeights);*/
		
		for (i = 0; i < numWeights; i++) {
			kernel1CovWeights[i] = 0.0;
			for (j = 0; j < numWeights; j++) {
				kernel1CovWeights[i] += kernel1Weights[j]*sysCovMatrix[covIndex[i][j].rowI + segIStart][covIndex[i][j].colI];
			}
		}
		ProfileS[sampI] = 0.0;
		for (i = 0; i < numWeights; i++) {
			ProfileS[sampI] += kernel1CovWeights[i]*kernel2Weights[i];
		}
		
	}
	
	/*Free memory */
	
	free(kernel1CovWeights);
	free(kernel2Weights);
	free(kernel1Weights);
	
	for (i = 0; i < numCovCols; i++) {
		free(covIndex[i]);
	}
	free(covIndex);
	
	freeDoubleMatrix(numCovRows, sysCovMatrix);
	free(aggSegI);
    
    return;
}


index2Dtype **makeMatrixIndex(int matrixLen){
	int i, j, coor1, coor2;
	index2Dtype **covIndex;
	
	covIndex = (index2Dtype **)malloc(sizeof(index2Dtype *) * matrixLen);
	for (i = 0; i < matrixLen; i++) {
		covIndex[i] = (index2Dtype *)malloc(sizeof(index2Dtype) * matrixLen);
		for (j = 0; j < matrixLen; j++) {
			coor1 = min(i, j);
			coor2 = max(i, j);
			covIndex[i][j].rowI = coor1;
			covIndex[i][j].colI = coor2 - coor1;
		}
	}
	return covIndex;
}

int getMaxCover(int numSamps, double *probeRange){
	int i, maxCover;
	
	maxCover = 0;
	for (i = 0; i < numSamps; i++) {
		maxCover = max(maxCover, probeRange[2*i+1] - probeRange[2*i]);
	}
	maxCover = maxCover - 1;
	return maxCover;
}


int *mexDoubleVector2IntVector(int numElem, double *vector){
	int i;
	int *intVector;
	
	intVector = (int *)malloc(sizeof(int) * numElem);
	for (i = 0; i <numElem; i++) {
		intVector[i] = (int)vector[i];
	}
	return intVector;
}

double **mexDoubleMatrix2DoubleMatrix(int numRows, int numCols, double *mexMatrix){
	int i, j;
	double **doubleMatrix;
	
	doubleMatrix = (double **)malloc(sizeof(double *) * numRows);
	for (i = 0; i < numRows; i++) {
		doubleMatrix[i] = (double *)malloc(sizeof(double) * numCols);
		for (j = 0; j < numCols; j++) {
			doubleMatrix[i][j] = mexMatrix[j*numRows + i];
		}
	}
	return doubleMatrix;
}

void freeDoubleMatrix(int numRows, double **mexMatrix){
	int i;
	for (i = 0; i < numRows; i++) {
		free(mexMatrix[i]);
	}
	free(mexMatrix);
}

int rnd(double number)
{
  return (number >= 0) ? (int)(number + 0.5) : (int)(number - 0.5);
}













