function [data, sigLayers,  sigLayersSign, sigLayersMu, sigLayersStd ] = detectModulesFDR( data, fieldParams, convInfo, conv, FDR, alpha)
    if length(FDR) > 1
        prevThresh = FDR(2);
        FDR = FDR(1)
    else
        prevThresh = FDR;
    end
    EE = FDR;
    numModules = 0;
    printf(-1,1,'FDR control:\n');
    iterI = 1;
    while 1
        printf(-1,1,'\tIter %d,  Number of events (previously) = %d, Allowed false positive rate = %e\n', iterI, numModules, EE);
        [data, sigLayers,  sigLayersSign, sigLayersMu, sigLayersStd ] = detectModules(data, fieldParams, convInfo, conv, EE, 0, alpha);
        numModules = 0;
        for i = 1:length(sigLayersSign)
            numModules = numModules + length(sigLayersSign{i});
        end

        EE = FDR*(numModules - EE) /(1-FDR);
        if EE <= prevThresh
            break;
        else
            prevThresh = EE;
        end
        
        iterI = iterI + 1;
        
    end
end



function [data, sigLayers,  sigLayersSign, sigLayersMu, sigLayersStd] = detectModules(data, fieldParams, convInfo, conv, EE, splitSign, alpha)

    currLayer = 1;
    printf(-1,1,'\t\tRecusive level 1 ... ');
    sigLayers = {};
    sigLayersSign = {};
    sigLayersMu = {};
    sigLayersStd = {};
    [data, segsPos, segsNeg, fieldParams] = getSigRegions(data, [1; length(convInfo.probeLoc)], fieldParams, convInfo, conv, EE, 0, alpha);
    [segsPos, segsNeg] = removeOverlapSegs(segsPos, segsNeg);
    
    if splitSign == 1
        sigLayers{1} = segsPos;
    elseif splitSign == -1
        sigLayers{1} = segsNeg;
    else
        sigLayers{1} = cat(2, segsPos, segsNeg);
    end
    [~, numPos] = size(segsPos);
    [~, numNeg] = size(segsNeg);

    if splitSign == 1
        sigLayersSign{1} = zeros(1,numPos);
        sigLayersSign{1}(1:numPos) = 1;
        sigLayersMu{1} = fieldParams.mu * ones(1,numPos);
        sigLayersStd{1} = fieldParams.sig * ones(1,numPos);
    elseif splitSign == -1
        sigLayersSign{1} = zeros(1,numNeg);
        sigLayersSign{1}(1:numNeg) = -1;
        sigLayersMu{1} = fieldParams.mu * ones(1,numNeg);
        sigLayersStd{1} = fieldParams.sig * ones(1,numNeg);
    else
        sigLayersSign{1} = zeros(1,numPos+numNeg);
        sigLayersSign{1}(1:numPos) = 1;
        sigLayersSign{1}(numPos+1:numPos+numNeg) = -1;
        sigLayersMu{1} = fieldParams.mu * ones(1,numPos+numNeg);
        sigLayersStd{1} = fieldParams.sig * ones(1,numPos+numNeg);
    end
    printf(-1,1,'100%%\n');

    while ~isempty(sigLayersSign{currLayer})
        [~, numSegs] = size(sigLayers{currLayer});
        currLayer = currLayer + 1;
        printf(-1,1,'\t\tRecusive level %d ... ', currLayer);
        printf(-1,1,'');
        sigLayers{currLayer} = [];
        sigLayersSign{currLayer} = [];
        sigLayersMu{currLayer} = [];
        sigLayersStd{currLayer} = [];
        for i = 1:numSegs
            if (sigLayers{currLayer-1}(2,i) - sigLayers{currLayer-1}(1,i) + 1) >= 12
                fieldParams.sig = sigLayersStd{currLayer-1}(i);
                [data, segsPos, segsNeg, fieldParams] = getSigRegions(data, sigLayers{currLayer-1}(:,i), fieldParams, convInfo, conv, EE, sigLayersSign{currLayer-1}(i), alpha);
                [segsPos, segsNeg] = removeOverlapSegs(segsPos, segsNeg);
                [~, numPos] = size(segsPos);
                [~, numNeg] = size(segsNeg);

                if splitSign == 1
                    sigLayers{currLayer} = cat(2, sigLayers{currLayer}, segsPos);
                    sigLayersSign{currLayer} = cat(2,sigLayersSign{currLayer}, +ones(1,numPos));
                    sigLayersMu{currLayer} = cat(2, sigLayersMu{currLayer}, fieldParams.mu * ones(1, numPos));
                    sigLayersStd{currLayer}  = cat(2, sigLayersStd{currLayer}, fieldParams.sig * ones(1,numPos));
                elseif splitSign == -1
                    sigLayers{currLayer} = cat(2, sigLayers{currLayer}, segsNeg);
                    sigLayersSign{currLayer} = cat(2, sigLayersSign{currLayer}, -ones(1,numNeg));
                    sigLayersMu{currLayer} = cat(2, sigLayersMu{currLayer}, fieldParams.mu * ones(1, numNeg));
                    sigLayersStd{currLayer}  = cat(2, sigLayersStd{currLayer}, fieldParams.sig * ones(1,numNeg));
                else
                    sigLayers{currLayer} = cat(2, sigLayers{currLayer}, segsPos);
                    sigLayers{currLayer} = cat(2, sigLayers{currLayer}, segsNeg);
                    sigLayersSign{currLayer} = cat(2,sigLayersSign{currLayer}, +ones(1,numPos));
                    sigLayersSign{currLayer} = cat(2, sigLayersSign{currLayer}, -ones(1,numNeg));
                    sigLayersMu{currLayer} = cat(2, sigLayersMu{currLayer}, fieldParams.mu * ones(1, numPos+numNeg));
                    sigLayersStd{currLayer}  = cat(2, sigLayersStd{currLayer}, fieldParams.sig * ones(1,numPos+numNeg));
                end
            end
            printf(1,1,'%0.2f%%', 100*i / numSegs);
        end
        printf(1,1,'100%%\n');
    end

end

function [segsPos, segsNeg] = removeOverlapSegs(segsPos, segsNeg)
    if (~isempty(segsPos)) && (~isempty(segsNeg))
        for segI = 1:length(segsPos(1,:))
            segRemI = (segsNeg(1,:) > segsPos(1,segI)) & (segsNeg(2,:) < segsPos(2,segI));
            segsNeg(:,segRemI) = [];
        end
        if ~isempty(segsNeg)
            for segI = 1:length(segsNeg(1,:))
                segRemI = (segsPos(1,:) > segsNeg(1,segI)) & (segsPos(2,:) < segsNeg(2,segI));
                segsPos(:,segRemI) = [];
            end
        end
    end
end


function [data, segsPos, segsNeg, fieldParams] = getSigRegions(data, seg,  fieldParams, convInfo, conv, thresh, prevSign, alpha)
    

    if seg(2,1) - seg(1,1) + 1 < 5
        segsPos = [];
        segsNeg = [];
        return;
    end

    lrAgr = data.lrAgr;
    nullRegionPrev = [];
    nullRegion = seg(1,1):seg(2,1);
%     if seg(2,1) - seg(1,1) + 1 >= 10*sum(fieldParams.R > 0.2)
    if seg(2,1) - seg(1,1) + 1 >= 200
        estStdFlag = 1;
    else
        estStdFlag = 0;
    end
    overlap = 0;
    nullRegionBuffer = {};

    while overlap < 0.99
        if  (prevSign ~= 0) || (overlap ~= 0)
            fieldParams.mu = mean(lrAgr(nullRegion));
        end
        [data, variance] = getParams(data, [], [], [], nullRegion, 'var');
        sig = sqrt(variance);        
        if (estStdFlag == 1)
            fieldParams.sig = max(sig, 0.1);
        else
            fieldParams.sig = max(sig, fieldParams.sig);
        end
        
        %  Incorperate systematic variation into the model
%         fieldParamsSys = fieldParams;
%         fieldParamsSys.R = fieldParamsSys.sig^2 * fieldParamsSys.R;
%         fieldParamsSys.R(1) = fieldParamsSys.R(1) + fieldParamsSys.sigSys^2;
%         fieldParamsSys.sig = sqrt(fieldParamsSys.R(1));
%         fieldParamsSys.R = fieldParamsSys.R / fieldParamsSys.R(1);
%         
        % Update scale space with field parameters
%         conv = updateScaleSpace(fieldParamsSys, conv);
        conv = updateScaleSpace(seg, fieldParams, convInfo, conv);
        % Get gain segments
        segsPos = getInfluenceSegs(data, seg, convInfo, conv, thresh, +1, alpha);

        % Get loss segments
        segsNeg = getInfluenceSegs(data, seg, convInfo, conv, thresh, -1, alpha);

        posRegion = segs2probes(segsPos);
        negRegion = segs2probes(segsNeg);
        sigRegion = union(posRegion, negRegion);
        nullRegionPrev = nullRegion;
        nullRegion = setdiff(seg(1,1):seg(2,1), sigRegion);
        if isempty(nullRegion)
            printf(-1,1,'Null region empty');
            if prevSign == 0
%                 nullRegionPrev = nullRegion;
%                 fieldParams
%                 plot(data.lrAgr)
                if length(posRegion) >= length(negRegion)
                    nullRegion = posRegion;
                else
                    nullRegion = negRegion;
                end
                
            elseif prevSign == 1
                nullRegion = posRegion;
            else
                nullRegion = negRegion;
            end
        end
        fieldParams.nullSegs = probes2segs(nullRegion);
        infLoopDetected = false;
        for i = 1:length(nullRegionBuffer)-1
            infLoopDetected = infLoopDetected || isequal(fieldParams.nullSegs, nullRegionBuffer{i});
        end
        if infLoopDetected
            break;
        end
        nullRegionBuffer{end+1} =  fieldParams.nullSegs;
        
        if isempty(nullRegion)  && (overlap == 0)
            overlap = 0.01;
            nullRegion = seg(1,1):seg(2,1);
        elseif ~isempty(union(nullRegionPrev, nullRegion))
            overlap = length(intersect(nullRegionPrev, nullRegion)) / length(union(nullRegionPrev, nullRegion));
        else
            overlap = 1;
        end
    end

    if prevSign == +1
        [~,numNegSegs] = size(segsNeg);
        if numNegSegs > 0
            if segsNeg(1,1) <= seg(1,1)
                segsNeg(:,1) = [];
            end
            if ~isempty(segsNeg)
                if segsNeg(2,end) >= seg(2,1)
                    segsNeg(:,end) = [];
                end
            end

        end
    end

    if prevSign == -1
        [~,numPosSegs] = size(segsPos);
        if numPosSegs > 0
            if segsPos(1,1) <= seg(1,1)
                segsPos(:,1) = [];
            end
            if ~isempty(segsPos)
                if segsPos(2,end) >= seg(2,1)
                    segsPos(:,end) = [];
                end
            end
        end
    end


    if (isequal(seg, segsPos) || isequal(seg, segsNeg))
        segsPos = [];
        segsNeg = [];
    end
end


function [conv] = updateScaleSpace(seg, fieldParams, convInfo, conv)
    for kwI = 1:length(conv.kw)
        sampStart = find(convInfo.kw{kwI}.mapLocSqsh < convInfo.kw{kwI}.probeLocSqsh(seg(1,1)), 1, 'last');
        sampEnd = find(convInfo.kw{kwI}.mapLocSqsh > convInfo.kw{kwI}.probeLocSqsh(seg(2,1)), 1, 'first');
        dK_C_dK = fieldParams.sig^2*conv.kw{kwI}.dK_R_dK + fieldParams.sigSys^2*conv.kw{kwI}.sys_dK_R_dK;
        K_C_dK   = fieldParams.sig^2*conv.kw{kwI}.K_R_dK + fieldParams.sigSys^2*conv.kw{kwI}.sys_K_R_dK;
        K_C_K     = fieldParams.sig^2*conv.kw{kwI}.K_R_K + fieldParams.sigSys^2*conv.kw{kwI}.sys_K_R_K;
        
        conv.kw{kwI}.zSmooth = (conv.kw{kwI}.lrSmooth - fieldParams.mu*conv.kw{kwI}.muSmooth)./(sqrt(K_C_K));
        conv.kw{kwI}.V = dK_C_dK./ K_C_K - (K_C_dK ./ K_C_K).^2;
        conv.kw{kwI}.V = max(conv.kw{kwI}.V, 0);
        conv.kw{kwI}.numEuler = 1;
        conv.kw{kwI}.intSqrtV    = sum(sqrt(conv.kw{kwI}.V)) * (convInfo.kw{kwI}.mapLocSqsh(2) - convInfo.kw{kwI}.mapLocSqsh(1));

        conv.kw{kwI}.eEuler = getAnalNegLogEEuler(conv.kw{kwI}.intSqrtV, conv.kw{kwI}.numEuler, abs(conv.kw{kwI}.zSmooth)) - log10(2);
%         if kwI == 1
%             figure;
%             plot(conv.kw{kwI}.eEuler)
%         end
    end
end


function [ returnSegs ] = getInfluenceSegs(data,  seg, convInfo, conv, thresh, tailSign, alpha)

    segLen = convInfo.probeLoc(seg(2,1)) - convInfo.probeLoc(seg(1,1));
    numKernels = sum(convInfo.kws <= segLen / alpha);
    minSig = -1000;

    % Set uninfluenced space to the minimum significance
    for kwI = 1:numKernels
        eulerMin = getAnalNegLogEEuler(conv.kw{kwI}.intSqrtV, conv.kw{kwI}.numEuler, 0) - log10(2);
        conv.kw{kwI}.eEuler(conv.kw{kwI}.eEuler < (eulerMin +1)) = minSig;
        conv.kw{kwI}.eEuler(sign(conv.kw{kwI}.zSmooth) == -tailSign) = minSig;
        
    end

    % Get sig segments in scalespace and delete wFactor widths  from the sides

    probesCover = [];
    for kwI = 1:numKernels
        sigIndice = find(conv.kw{kwI}.eEuler >= -log10(thresh));
        sigSegs = probes2segs(sigIndice);
        sigSegs = splitSegsOnChromBound(sigSegs, data, convInfo.kw{kwI}.probeLocSqsh, convInfo.kw{kwI}.mapLocSqsh);
        sigProbeSegs = convInfo.kw{kwI}.nearestProbeI(sigSegs')';
        if ~isempty(sigProbeSegs)
            probeLeftIn    =sigProbeSegs(1,:);
            probeRightIn = sigProbeSegs(2,:);
            sigSegSizes = convInfo.kw{kwI}.probeLocSqsh(probeRightIn) - convInfo.kw{kwI}.probeLocSqsh(probeLeftIn);
            sizeTooSmall = sigSegSizes < alpha*convInfo.kws(kwI);
            if convInfo.kws(kwI) ~= min(convInfo.kws);
                sigSegs(:,sizeTooSmall) = [];
            end
        end
        sigProbeSegs = convInfo.kw{kwI}.nearestProbeI(sigSegs')';
        sigProbes = segs2probes(sigProbeSegs);
        probesCover =  union(probesCover, sigProbes);
    end
    probesCover = intersect(probesCover, seg(1,1):seg(2,1));
    probesCover;

    segsCover = [];
    chroms = unique(data.chrom);
    for chrom = chroms
        currChromProbes = find(data.chrom == chrom);
        currProbes = intersect(currChromProbes, probesCover);
        if  ~isempty(currProbes)
            segsCover = cat(2, segsCover, probes2segs(currProbes));
        end   
    end
    returnSegs = segsCover;
end

function segs = splitSegsOnChromBound(segs, data, probeLoc, mapLoc)
    if ~isempty(segs)
        chromEndProbesI = find(diff(data.chrom) ~= 0);
        chromStartProbesI = chromEndProbesI + 1;
        chromEndProbeLocs = probeLoc(chromEndProbesI);
        chromStartProbeLocs = probeLoc(chromStartProbesI);
        chromBoundLocs = (chromEndProbeLocs + chromStartProbeLocs) / 2;
        for chromI = 1:length(chromBoundLocs)
            leftI    = find(mapLoc <= chromBoundLocs(chromI), 1, 'last');
            rightI = find(mapLoc >= chromBoundLocs(chromI), 1, 'first');

            overlapI = find((mapLoc(segs(1,:)) <= chromBoundLocs(chromI)) & (mapLoc(segs(2,:)) >= chromBoundLocs(chromI)));
            newSegs = segs(:,overlapI);
            newSegs(1,:) = rightI;
            segs(2,overlapI) = leftI;
            segs = cat(2, segs, newSegs);
            [~,segI] = sort(segs(1,:), 'ascend');
            segs = segs(:,segI);
        end
    end
end

function [segments] = probes2segs(probes)
    if isempty(probes)
        segments = [];
        return;
    end
    probesI = zeros(1, max(probes)+2);
    probesI(probes+1) = 1;
    probesDiff = diff(probesI);
    segments(1,:) = find(probesDiff == 1) + 1;
    segments(2,:) = find(probesDiff == -1);
    segments = segments - 1;
end

function [ probes ] = segs2probes( segments )
    probes = [];
    if isempty(segments)
        return;
    end
    numSegments = length(segments(1,:));
    for i = 1:numSegments
        probes = cat(2,probes, segments(1,i):segments(2,i));
    end
    probes = unique(probes);
end



